#!/usr/bin/env perl
use strict;
use warnings;

use utf8;
use Path::Class qw/file/;
use Perl6::Say;
use Encode;
use IO::Handle;
use Term::Encoding qw/term_encoding/;

my %index;
my %df;

my $fh = file('./dict_sample.tsv')->openr;
while (local $_ = $fh->getline) {
    chomp;

    my ($term, $df) = split /\t/;
    $term = decode_utf8 $term;

    add_to_index(\%index, $term);
    $df{$term} = sprintf "%d", $df;
}

while (1) {
    printf STDERR '> ';

    my $q = decode(term_encoding, STDIN->getline);
    chomp $q;

    my @cand = search_index(\%index, $q);
    my %score = map { $_ => edit_distance($q, $_) } @cand;
    for my $term ( sort { ($score{$a} <=> $score{$b}) || ($df{$b} <=> $df{$a}) } @cand ) {
        say sprintf "%s => %d", $term, $score{$term};
    }
}

sub add_to_index {
    my ($index, $term) = @_;
    for my $bi (ngram($term, 2)) {
        if (exists $index->{$bi}) {
            push @{$index->{$bi}}, $term;
        } else {
            $index->{$bi} = [ $term ];
        }
    }
}

sub search_index {
    my ($index, $q) = @_;
    my @res;
    my %occ;
    for my $bi (ngram($q, 2)) {
        if (my $terms = $index->{$bi}) {
            for my $t (@$terms) {
                if (++$occ{$t} == 2) {
                    push @res, $t;
                }
            }
        }
    }
    return @res;
}

##

use utf8;
use List::Util qw/min/;

sub ngram {
    my ($str, $n) = @_;
    $n ||= 2;
    my $len = length $str;

    if ($len < $n) {
        return $str;
    }

    my @ngrams;
    for my $i (0..$len - $n) {
        push @ngrams, substr($str, $i, $n);
    }
    return @ngrams;
}

sub edit_distance {
    my ($s1, $s2) = @_;
    my @m;
    my @s1 = split //, $s1;
    my @s2 = split //, $s2;

    for (my $i = 0; $i <= @s1; $i++) {
        $m[$i][0] = $i;
    }

    for (my $j = 0; $j <= @s2; $j++) {
        $m[0][$j] = $j;
    }

    for (my $i = 1; $i <= @s1; $i++) {
        for (my $j = 1; $j <= @s2; $j++) {
            my $diff = ($s1[$i - 1] eq $s2[$j - 1]) ? 0 : 1;
            $m[$i][$j] = min(
                $m[$i - 1][$j - 1] + $diff,
                $m[$i - 1][$j] + 1,
                $m[$i][$j - 1] + 1,
            );
        }
    }

    return $m[-1][-1];
}
